@@ -30,16 +30,6 @@
 #include <glib/gthread.h>
 #include <gtk/gtk.h>
 
-#if HAVE_EXIF
-#include <libexif/exif-data.h>
-#include <libexif/exif-ifd.h>
-#include <libexif/exif-content.h>
-#include <libexif/exif-tag.h>
-#include <libexif/exif-byte-order.h>
-#include <libexif/exif-format.h>
-#include <libexif/exif-utils.h>
-#endif /* HAVE_EXIF */
-
 #include "gs-theme-engine.h"
 #include "gste-slideshow.h"
 
@@ -77,8 +67,11 @@ struct GSTESlideshowPrivate
 
         GSList         *filename_list;
         char           *images_location;
+        gboolean	sort_images;
         int             window_width;
         int             window_height;
+        PangoColor     *background_color;
+        gboolean        no_stretch_hint;
 
         guint           timeout_id;
 
@@ -88,7 +81,10 @@ struct GSTESlideshowPrivate
 
 enum {
         PROP_0,
-        PROP_IMAGES_LOCATION
+        PROP_IMAGES_LOCATION,
+        PROP_SORT_IMAGES,
+        PROP_SOLID_BACKGROUND,
+        PROP_NO_STRETCH_HINT
 };
 
 #define GSTE_SLIDESHOW_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GSTE_TYPE_SLIDESHOW, GSTESlideshowPrivate))
@@ -113,231 +109,6 @@ typedef struct _OpResult
         GdkPixbuf *pixbuf;
 } OpResult;
 
-#if HAVE_EXIF
-
-enum _ExifRotation {
-        NORMAL = 1,
-        HFLIP,
-        ROTATE180,
-        VFLIP,
-        ROTATE90_HFLIP,
-        ROTATE90,
-        ROTATE90_VFLIP,
-        ROTATE270
-};
-typedef enum _ExifRotation ExifRotation;
-
-static ExifRotation
-get_exif_orientation (const char *filename)
-{
-        ExifData     *edata;
-        ExifEntry    *entry;
-        ExifByteOrder o;
-        ExifRotation  s;
-        unsigned int  i;
-
-        edata = exif_data_new_from_file (filename);
-
-        if (edata == NULL) {
-                return 0;
-        }
-
-        o = exif_data_get_byte_order (edata);
-
-        for (i = 0; i < EXIF_IFD_COUNT; i++) {
-                entry = exif_content_get_entry (edata->ifd [i],
-                                                EXIF_TAG_ORIENTATION);
-                if (entry != NULL) {
-                        s = (ExifRotation)exif_get_short (entry->data, o);
-                        exif_data_unref (edata);
-
-                        return s;
-                }
-        }
-
-        exif_data_unref (edata);
-
-        return 0;
-}
-
-/* Routines for rotating a GdkPixbuf.
- * Borrowed from f-spot.
- * For copyright information see libfspot/f-pixbuf-utils.c in the f-spot
- * package.
- */
-
-/* Returns a copy of pixbuf src rotated 90 degrees clockwise or 90
-   counterclockwise.  */
-static GdkPixbuf *
-pixbuf_copy_rotate_90 (GdkPixbuf *src,
-                       gboolean   counter_clockwise)
-{
-        GdkPixbuf *dest;
-        int        has_alpha;
-        int        sw, sh, srs;
-        int        dw, dh, drs;
-        guchar    *s_pix;
-        guchar    *d_pix;
-        guchar    *sp;
-        guchar    *dp;
-        int        i, j;
-        int        a;
-
-        if (src == NULL) {
-                return NULL;
-        }
-
-        sw = gdk_pixbuf_get_width (src);
-        sh = gdk_pixbuf_get_height (src);
-        has_alpha = gdk_pixbuf_get_has_alpha (src);
-        srs = gdk_pixbuf_get_rowstride (src);
-        s_pix = gdk_pixbuf_get_pixels (src);
-
-        dw = sh;
-        dh = sw;
-        dest = gdk_pixbuf_new (GDK_COLORSPACE_RGB, has_alpha, 8, dw, dh);
-        drs = gdk_pixbuf_get_rowstride (dest);
-        d_pix = gdk_pixbuf_get_pixels (dest);
-
-        a = (has_alpha ? 4 : 3);
-
-        for (i = 0; i < sh; i++) {
-                sp = s_pix + (i * srs);
-                for (j = 0; j < sw; j++) {
-                        if (counter_clockwise) {
-                                dp = d_pix + ((dh - j - 1) * drs) + (i * a);
-                        } else {
-                                dp = d_pix + (j * drs) + ((dw - i - 1) * a);
-                        }
-
-                        *(dp++) = *(sp++);	/* r */
-                        *(dp++) = *(sp++);	/* g */
-                        *(dp++) = *(sp++);	/* b */
-                        if (has_alpha) {
-                                *(dp) = *(sp++);	/* a */
-                        }
-                }
-        }
-
-        return dest;
-}
-
-/* Returns a copy of pixbuf mirrored and or flipped.  TO do a 180 degree
-   rotations set both mirror and flipped TRUE if mirror and flip are FALSE,
-   result is a simple copy.  */
-static GdkPixbuf *
-pixbuf_copy_mirror (GdkPixbuf *src,
-                    gboolean   mirror,
-                    gboolean   flip)
-{
-        GdkPixbuf *dest;
-        int        has_alpha;
-        int        w, h, srs;
-        int        drs;
-        guchar    *s_pix;
-        guchar    *d_pix;
-        guchar    *sp;
-        guchar    *dp;
-        int        i, j;
-        int        a;
-
-        if (src == NULL) {
-                return NULL;
-        }
-
-        w = gdk_pixbuf_get_width (src);
-        h = gdk_pixbuf_get_height (src);
-        has_alpha = gdk_pixbuf_get_has_alpha (src);
-        srs = gdk_pixbuf_get_rowstride (src);
-        s_pix = gdk_pixbuf_get_pixels (src);
-
-        dest = gdk_pixbuf_new (GDK_COLORSPACE_RGB, has_alpha, 8, w, h);
-        drs = gdk_pixbuf_get_rowstride (dest);
-        d_pix = gdk_pixbuf_get_pixels (dest);
-
-        a = has_alpha ? 4 : 3;
-
-        for (i = 0; i < h; i++)	{
-                sp = s_pix + (i * srs);
-                if (flip) {
-                        dp = d_pix + ((h - i - 1) * drs);
-                } else {
-                        dp = d_pix + (i * drs);
-                }
-
-                if (mirror) {
-                        dp += (w - 1) * a;
-                        for (j = 0; j < w; j++) {
-                                *(dp++) = *(sp++);	/* r */
-                                *(dp++) = *(sp++);	/* g */
-                                *(dp++) = *(sp++);	/* b */
-                                if (has_alpha) {
-                                        *(dp) = *(sp++);	/* a */
-                                }
-                                dp -= (a + 3);
-                        }
-                } else {
-                        for (j = 0; j < w; j++) {
-                                *(dp++) = *(sp++);	/* r */
-                                *(dp++) = *(sp++);	/* g */
-                                *(dp++) = *(sp++);	/* b */
-                                if (has_alpha) {
-                                        *(dp++) = *(sp++);	/* a */
-                                }
-                        }
-                }
-        }
-
-        return dest;
-}
-
-static GdkPixbuf *
-update_from_exif_data (char      *filename,
-                       GdkPixbuf *pixbuf)
-{
-        GdkPixbuf *newpixbuf = NULL;
-        GdkPixbuf *tmppixbuf = NULL;
-
-        if (pixbuf == NULL) {
-                return NULL;
-        }
-
-        switch (get_exif_orientation (filename)) {
-        case NORMAL:
-                break;
-        case HFLIP:
-                newpixbuf = pixbuf_copy_mirror (pixbuf, TRUE, FALSE);
-                break;
-        case ROTATE180:
-                newpixbuf = pixbuf_copy_mirror (pixbuf, TRUE, TRUE);
-                break;
-        case VFLIP:
-                newpixbuf = pixbuf_copy_mirror (pixbuf, FALSE, TRUE);
-                break;
-        case ROTATE90_HFLIP:
-                tmppixbuf = pixbuf_copy_mirror (pixbuf, FALSE, TRUE);
-                newpixbuf = pixbuf_copy_rotate_90 (tmppixbuf, FALSE);
-                g_object_unref (tmppixbuf);
-                break;
-        case ROTATE90:
-                newpixbuf = pixbuf_copy_rotate_90 (pixbuf, FALSE);
-                break;
-        case ROTATE90_VFLIP:
-                tmppixbuf = pixbuf_copy_mirror (pixbuf, TRUE, FALSE);
-                newpixbuf = pixbuf_copy_rotate_90 (tmppixbuf, FALSE);
-                g_object_unref (tmppixbuf);
-                break;
-        case ROTATE270:
-                newpixbuf = pixbuf_copy_rotate_90 (pixbuf, TRUE);
-                break;
-        default:
-                break;
-        }
-
-        return newpixbuf;
-}
-#endif /* HAVE_EXIF */
-
 static gboolean
 push_load_image_func (GSTESlideshow *show)
 {
@@ -397,6 +168,31 @@ start_fade (GSTESlideshow *show,
         x = (window_width - pw) / 2;
         y = (window_height - ph) / 2;
 
+        if (gdk_pixbuf_get_has_alpha (pixbuf) && show->priv->background_color) {
+                GdkPixbuf *colored;
+                guint32    color;
+                GdkPixmap *pixmap;
+
+                color = (show->priv->background_color->red << 16)
+                        + (show->priv->background_color->green / 256 << 8)
+                        + show->priv->background_color->blue / 256;
+                colored = gdk_pixbuf_composite_color_simple (pixbuf,
+                                                             pw, ph,
+                                                             GDK_INTERP_BILINEAR,
+                                                             255,
+                                                             256,
+                                                             color,
+                                                             color);
+                pixmap = gdk_pixmap_new (NULL, ph, pw,  gdk_visual_get_system ()->depth);
+
+                gdk_draw_pixbuf (pixmap, NULL, colored, 0, 0, 0, 0, -1, -1, GDK_RGB_DITHER_MAX, 0, 0);
+                gdk_pixbuf_get_from_drawable (pixbuf, pixmap, NULL, 0, 0, 0, 0, -1, -1);
+
+                g_object_unref (pixmap);
+
+                g_object_unref(colored);
+        }
+
         cr = cairo_create (show->priv->surf);
 
         /* XXX Handle out of memory? */
@@ -451,25 +247,49 @@ update_display (GSTESlideshow *show)
                 /* fade out areas not covered by the new image */
                 /* top */
                 cairo_rectangle (cr, 0, 0, window_width, show->priv->pat2top);
-                cairo_set_source_rgba (cr, 0.0, 0.0, 0.0, show->priv->alpha2);
+                if (show->priv->background_color) {
+                        cairo_set_source_rgba (cr, show->priv->background_color->red / 65535.0,
+                               show->priv->background_color->green / 65535.0,
+                               show->priv->background_color->blue / 65535.0, show->priv->alpha2);
+                } else {
+	                cairo_set_source_rgba (cr, 0.0, 0.0, 0.0, show->priv->alpha2);
+                }
                 cairo_fill (cr);
                 /* left (excluding what's covered by top and bottom) */
                 cairo_rectangle (cr, 0, show->priv->pat2top,
                                  show->priv->pat2left,
                                  show->priv->pat2bottom - show->priv->pat2top);
-                cairo_set_source_rgba (cr, 0.0, 0.0, 0.0, show->priv->alpha2);
+                if (show->priv->background_color) {
+                        cairo_set_source_rgba (cr, show->priv->background_color->red / 65535.0,
+                               show->priv->background_color->green / 65535.0,
+                               show->priv->background_color->blue / 65535.0, show->priv->alpha2);
+                } else {
+	                cairo_set_source_rgba (cr, 0.0, 0.0, 0.0, show->priv->alpha2);
+                }
                 cairo_fill (cr);
                 /* bottom */
                 cairo_rectangle (cr, 0, show->priv->pat2bottom, window_width,
                                  window_height - show->priv->pat2bottom);
-                cairo_set_source_rgba (cr, 0.0, 0.0, 0.0, show->priv->alpha2);
+                if (show->priv->background_color) {
+                        cairo_set_source_rgba (cr, show->priv->background_color->red / 65535.0,
+                               show->priv->background_color->green / 65535.0,
+                               show->priv->background_color->blue / 65535.0, show->priv->alpha2);
+                } else {
+	                cairo_set_source_rgba (cr, 0.0, 0.0, 0.0, show->priv->alpha2);
+                }
                 cairo_fill (cr);
                 /* right (excluding what's covered by top and bottom) */
                 cairo_rectangle (cr, show->priv->pat2right,
                                  show->priv->pat2top,
                                  window_width - show->priv->pat2right,
                                  show->priv->pat2bottom - show->priv->pat2top);
-                cairo_set_source_rgba (cr, 0.0, 0.0, 0.0, show->priv->alpha2);
+                if (show->priv->background_color)  {
+                        cairo_set_source_rgba (cr, show->priv->background_color->red / 65535.0,
+                               show->priv->background_color->green / 65535.0,
+                               show->priv->background_color->blue / 65535.0, show->priv->alpha2);
+                } else {
+	                cairo_set_source_rgba (cr, 0.0, 0.0, 0.0, show->priv->alpha2);
+                }
                 cairo_fill (cr);
 
                 gs_theme_engine_profile_start ("paint pattern to surface");
@@ -618,14 +438,14 @@ results_pull_func (GSTESlideshow *show)
 static GdkPixbuf *
 scale_pixbuf (GdkPixbuf *pixbuf,
               int        max_width,
-              int        max_height)
+              int        max_height,
+              gboolean   no_stretch_hint)
 {
         int        pw;
         int        ph;
         float      scale_factor_x = 1.0;
         float      scale_factor_y = 1.0;
         float      scale_factor = 1.0;
-        GdkPixbuf *scaled = NULL;
 
         pw = gdk_pixbuf_get_width (pixbuf);
         ph = gdk_pixbuf_get_height (pixbuf);
@@ -646,17 +466,18 @@ scale_pixbuf (GdkPixbuf *pixbuf,
                 scale_factor = scale_factor_x;
         }
 
-        if (1) {
+        /* always scale down, allow to disable scaling up */
+        if (scale_factor < 1.0 || !no_stretch_hint) {
                 int scale_x = (int) (pw * scale_factor);
                 int scale_y = (int) (ph * scale_factor);
 
-                scaled = gdk_pixbuf_scale_simple (pixbuf,
-                                                  scale_x,
-                                                  scale_y,
-                                                  GDK_INTERP_BILINEAR);
+                return gdk_pixbuf_scale_simple (pixbuf,
+                                                scale_x,
+                                                scale_y,
+                                                GDK_INTERP_BILINEAR);
+        } else {
+                return g_object_ref (pixbuf);
         }
-
-        return scaled;
 }
 
 static void
@@ -696,11 +517,19 @@ build_filename_list_local_dir (const char *base)
         return list;
 }
 
+static int
+gste_strcmp_compare_func (gconstpointer string_a, gconstpointer string_b)
+{
+        return strcmp (string_a == NULL ? "" : string_a,
+                       string_b == NULL ? "" : string_b);
+}
+
+
 static GdkPixbuf *
 get_pixbuf_from_local_dir (GSTESlideshow *show,
                            const char    *location)
 {
-        GdkPixbuf *pixbuf;
+        GdkPixbuf *pixbuf, *transformed_pixbuf;
         char      *filename;
         int        i;
         GSList    *l;
@@ -712,32 +541,34 @@ get_pixbuf_from_local_dir (GSTESlideshow *show,
 
         if (show->priv->filename_list == NULL) {
                 return NULL;
+        } else {
+                if (show->priv->sort_images) {
+                        show->priv->filename_list = g_slist_sort (show->priv->filename_list, gste_strcmp_compare_func);
+                }
         }
 
-        /* get a random filename */
-        i = g_random_int_range (0, g_slist_length (show->priv->filename_list));
-        l = g_slist_nth (show->priv->filename_list, i);
+        /* get a random filename if needed */
+        if (! show->priv->sort_images) {
+                i = g_random_int_range (0, g_slist_length (show->priv->filename_list));
+                l = g_slist_nth (show->priv->filename_list, i);
+        } else {
+                l = show->priv->filename_list;
+        }
         filename = l->data;
 
         pixbuf = gdk_pixbuf_new_from_file (filename, NULL);
 
-#if HAVE_EXIF
-        /* try to get exif data and rotate the image if necessary */
         if (pixbuf != NULL) {
-                GdkPixbuf *rotated;
-
-                rotated = update_from_exif_data (filename, pixbuf);
-                if (rotated) {
-                        g_object_unref (pixbuf);
-                        pixbuf = rotated;
-                }
+                transformed_pixbuf = gdk_pixbuf_apply_embedded_orientation (pixbuf);
+                g_object_unref (pixbuf);
+        } else {
+                transformed_pixbuf = NULL;
         }
-#endif /* HAVE_EXIF */
 
         g_free (filename);
         show->priv->filename_list = g_slist_delete_link (show->priv->filename_list, l);
 
-        return pixbuf;
+        return transformed_pixbuf;
 }
 
 static GdkPixbuf *
@@ -776,7 +607,7 @@ get_pixbuf (GSTESlideshow *show,
         pixbuf = get_pixbuf_from_location (show, location);
 
         if (pixbuf != NULL) {
-                scaled = scale_pixbuf (pixbuf, width, height);
+                scaled = scale_pixbuf (pixbuf, width, height, show->priv->no_stretch_hint);
                 g_object_unref (pixbuf);
         }
 
@@ -847,6 +678,46 @@ gste_slideshow_set_images_location (GSTESlideshow *show,
         show->priv->images_location = g_strdup (location);
 }
 
+
+void
+gste_slideshow_set_sort_images (GSTESlideshow *show,
+                                gboolean       sort_images)
+{
+        g_return_if_fail (GSTE_IS_SLIDESHOW (show));
+
+        show->priv->sort_images = sort_images;
+}
+
+void
+gste_slideshow_set_no_stretch_hint (GSTESlideshow *show,
+                                     gboolean       no_stretch_hint)
+{
+        g_return_if_fail (GSTE_IS_SLIDESHOW (show));
+
+        show->priv->no_stretch_hint = no_stretch_hint;
+}
+
+void
+gste_slideshow_set_background_color (GSTESlideshow *show,
+                                     const char    *background_color)
+{
+        g_return_if_fail (GSTE_IS_SLIDESHOW (show));
+
+        if (show->priv->background_color != NULL) {
+                g_slice_free (PangoColor, show->priv->background_color);
+                show->priv->background_color = NULL;
+        }
+
+        if (background_color != NULL) {
+               show->priv->background_color = g_slice_new (PangoColor);
+
+               if (pango_color_parse (show->priv->background_color, background_color) == FALSE) {
+                       g_slice_free (PangoColor, show->priv->background_color);
+                       show->priv->background_color = NULL;
+               }
+        }
+}
+
 static void
 gste_slideshow_set_property (GObject            *object,
                              guint               prop_id,
@@ -861,6 +732,15 @@ gste_slideshow_set_property (GObject            *object,
         case PROP_IMAGES_LOCATION:
                 gste_slideshow_set_images_location (self, g_value_get_string (value));
                 break;
+        case PROP_SORT_IMAGES:
+                gste_slideshow_set_sort_images (self, g_value_get_boolean (value));
+                break;
+        case PROP_SOLID_BACKGROUND:
+                gste_slideshow_set_background_color (self, g_value_get_string (value));
+                break;
+	case PROP_NO_STRETCH_HINT:
+		gste_slideshow_set_no_stretch_hint (self, g_value_get_boolean (value));
+		break;
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
                 break;
@@ -881,6 +761,20 @@ gste_slideshow_get_property (GObject            *object,
         case PROP_IMAGES_LOCATION:
                 g_value_set_string (value, self->priv->images_location);
                 break;
+	case PROP_SORT_IMAGES:
+		g_value_set_boolean (value, self->priv->sort_images);
+		break;
+        case PROP_SOLID_BACKGROUND:
+                {
+                        char *color = NULL;
+                        color = pango_color_to_string (self->priv->background_color);
+                        g_value_set_string (value, color);
+                        g_free (color);
+                        break;
+                }
+	case PROP_NO_STRETCH_HINT:
+		g_value_set_boolean (value, self->priv->no_stretch_hint);
+		break;
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
                 break;
@@ -987,7 +881,27 @@ gste_slideshow_class_init (GSTESlideshowClass *klass)
                                                               NULL,
                                                               NULL,
                                                               G_PARAM_READWRITE));
-
+        g_object_class_install_property (object_class,
+                                         PROP_SORT_IMAGES,
+                                         g_param_spec_boolean ("sort-images",
+                                                               NULL,
+                                                               NULL,
+                                                               FALSE,
+                                                               G_PARAM_READWRITE));
+        g_object_class_install_property (object_class,
+                                         PROP_SOLID_BACKGROUND,
+                                         g_param_spec_string ("background-color",
+                                                              NULL,
+                                                              NULL,
+                                                              NULL,
+                                                              G_PARAM_READWRITE));
+        g_object_class_install_property (object_class,
+                                         PROP_NO_STRETCH_HINT,
+                                         g_param_spec_boolean ("no-stretch",
+                                                               NULL,
+                                                               NULL,
+                                                               FALSE,
+                                                               G_PARAM_READWRITE));
 }
 
 static void
@@ -1067,6 +981,11 @@ gste_slideshow_finalize (GObject *object)
         g_free (show->priv->images_location);
         show->priv->images_location = NULL;
 
+	if (show->priv->background_color) {
+		g_slice_free (PangoColor, show->priv->background_color);
+		show->priv->background_color = NULL;
+	}
+
         if (show->priv->timer != NULL) {
                 g_timer_destroy (show->priv->timer);
         }
