@@ -833,6 +833,7 @@ submit_note (GtkButton  *button,
 #ifdef WITH_LIBNOTIFY
         char               *text;
         char                summary[128];
+        char               *escaped_text;
         GtkTextBuffer      *buffer;
         GtkTextIter         start, end;
         time_t              t;
@@ -844,18 +845,20 @@ submit_note (GtkButton  *button,
         gtk_text_buffer_get_bounds (buffer, &start, &end);
         text = gtk_text_buffer_get_text (buffer, &start, &end, FALSE);
         gtk_text_buffer_set_text (buffer, "", 0);
+        escaped_text = g_markup_escape_text (text, -1);
 
         t = time (NULL);
         tmp = localtime (&t);
         strftime (summary, 128, "%X", tmp);
 
         notify_init ("gnome-screensaver-dialog");
-        note = notify_notification_new (summary, text, NULL, NULL);
+        note = notify_notification_new (summary, escaped_text, NULL, NULL);
         notify_notification_set_timeout (note, NOTIFY_EXPIRES_NEVER);
         notify_notification_show (note, NULL);
         g_object_unref (note);
 
         g_free (text);
+        g_free (escaped_text);
 
         gs_lock_plug_response (plug, GS_LOCK_PLUG_RESPONSE_CANCEL);
 #endif /* WITH_LIBNOTIFY */
@@ -873,6 +876,8 @@ cancel_note (GtkButton  *button,
 
         /* this counts as activity so restart the timer */
         restart_cancel_timeout (plug);
+
+        gtk_window_set_default (GTK_WINDOW (plug), plug->priv->auth_unlock_button);
 }
 
 static void
@@ -1560,6 +1565,19 @@ delete_handler (GSLockPlug  *plug,
         return TRUE; /* Do not destroy */
 }
 
+static void
+clear_clipboards (GSLockPlug *plug)
+{
+        GtkClipboard *clipboard;
+
+        clipboard = gtk_widget_get_clipboard (GTK_WIDGET (plug), GDK_SELECTION_PRIMARY);
+        gtk_clipboard_clear (clipboard);
+        gtk_clipboard_set_text (clipboard, "", -1);
+        clipboard = gtk_widget_get_clipboard (GTK_WIDGET (plug), GDK_SELECTION_CLIPBOARD);
+        gtk_clipboard_clear (clipboard);
+        gtk_clipboard_set_text (clipboard, "", -1);
+}
+
 #define INVISIBLE_CHAR_DEFAULT       '*'
 #define INVISIBLE_CHAR_BLACK_CIRCLE  0x25cf
 #define INVISIBLE_CHAR_WHITE_BULLET  0x25e6
@@ -1575,6 +1593,8 @@ gs_lock_plug_init (GSLockPlug *plug)
 
         plug->priv = GS_LOCK_PLUG_GET_PRIVATE (plug);
 
+        clear_clipboards (plug);
+
 #ifdef WITH_LIBNOTIFY
         plug->priv->leave_note_enabled = TRUE;
 #else
